@@ -1,17 +1,17 @@
-# Copyright (C) 2012 Jason A. Donenfeld <Jason@zx2c4.com>. All Rights Reserved.
+# Copyright (C) 2012 - 2014 Jason A. Donenfeld <Jason@zx2c4.com>. All Rights Reserved.
 # This file is licensed under the GPLv2+. Please see COPYING for more information.
 
 clip() {
+	sleep_argv0="password store sleep for user $(id -u)"
+	pkill -f "^$sleep_argv0" 2>/dev/null && sleep 0.5
 	before="$(pbpaste | openssl base64)"
 	echo -n "$1" | pbcopy
 	(
-		sleep 45
+		( exec -a "$sleep_argv0" sleep "$CLIP_TIME" )
 		now="$(pbpaste | openssl base64)"
-		if [[ $now != $(echo -n "$1" | openssl base64) ]]; then
-			before="$now"
-		fi
+		[[ $now != $(echo -n "$1" | openssl base64) ]] && before="$now"
 		echo "$before" | openssl base64 -d | pbcopy
-	) & disown
+	) 2>/dev/null & disown
 	echo "Copied $2 to clipboard. Will clear in 45 seconds."
 }
 
@@ -32,3 +32,4 @@ tmpdir() {
 }
 
 GETOPT="$(brew --prefix gnu-getopt 2>/dev/null || echo /usr/local)/bin/getopt"
+SHRED="srm -f -z"
